@@ -44,6 +44,7 @@ void hitori_destroy_cb (GtkWindow *window, Hitori *hitori);
 void hitori_window_state_event_cb (GtkWindow *window, GdkEventWindowState *event, Hitori *hitori);
 static void new_game_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
 static void hint_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
+static void quit_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
 static void undo_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
 static void redo_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
 static void help_cb (GSimpleAction *action, GVariant *parameter, gpointer user_data);
@@ -54,6 +55,7 @@ static GActionEntry app_entries[] = {
 	{ "new-game", new_game_cb, NULL, NULL, NULL },
 	{ "about", about_cb, NULL, NULL, NULL },
 	{ "help", help_cb, NULL, NULL, NULL },
+	{ "quit", quit_cb, NULL, NULL, NULL },
 };
 
 static GActionEntry win_entries[] = {
@@ -62,6 +64,33 @@ static GActionEntry win_entries[] = {
 	{ "redo", redo_cb, NULL, NULL, NULL },
 };
 
+static void
+hitori_window_unmap_cb (GtkWidget *window,
+                        gpointer user_data)
+{
+	gboolean window_maximized;
+	GdkRectangle geometry;
+	HitoriApplication *hitori;
+
+	hitori = HITORI_APPLICATION (user_data);
+
+	window_maximized = gtk_window_is_maximized (GTK_WINDOW (window));
+	g_settings_set_boolean (hitori->settings,
+				"window-maximized", window_maximized);
+
+	if (window_maximized)
+		return;
+
+	gtk_window_get_position (GTK_WINDOW (window), &geometry.x, &geometry.y);
+	gtk_window_get_size (GTK_WINDOW (window),
+			     &geometry.width, &geometry.height);
+
+	g_settings_set (hitori->settings, "window-position", "(ii)",
+			geometry.x, geometry.y);
+	g_settings_set (hitori->settings, "window-size", "(ii)",
+			geometry.width, geometry.height);
+}
+
 GtkWidget *
 hitori_create_interface (Hitori *hitori)
 {
@@ -81,6 +110,9 @@ hitori_create_interface (Hitori *hitori)
 	hitori->drawing_area = GTK_WIDGET (gtk_builder_get_object (builder, "hitori_drawing_area"));
 	hitori->timer_label = GTK_LABEL (gtk_builder_get_object (builder, "hitori_timer"));
 
+	g_signal_connect (hitori->window, "unmap",
+			  G_CALLBACK (hitori_window_unmap_cb), hitori);
+
 	g_object_unref (builder);
 
 	/* Set up actions */
@@ -96,6 +128,20 @@ hitori_create_interface (Hitori *hitori)
 	hitori->redo_action = G_SIMPLE_ACTION (g_action_map_lookup_action (G_ACTION_MAP (hitori->window), "redo"));
 	hitori->hint_action = G_SIMPLE_ACTION (g_action_map_lookup_action (G_ACTION_MAP (hitori->window), "hint"));
 
+	const gchar *vaccels_help[] = {"F1", NULL};
+	const gchar *vaccels_hint[] = {"<Primary>h", NULL};
+	const gchar *vaccels_new[] = {"<Primary>n", NULL};
+	const gchar *vaccels_quit[] = {"<Primary>q", "<Primary>w", NULL};
+	const gchar *vaccels_redo[] = {"<Primary><Shift>z", NULL};
+	const gchar *vaccels_undo[] = {"<Primary>z", NULL};
+
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "app.help", vaccels_help);
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "win.hint", vaccels_hint);
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "app.new-game", vaccels_new);
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "app.quit", vaccels_quit);
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "win.redo", vaccels_redo);
+	gtk_application_set_accels_for_action (GTK_APPLICATION (hitori), "win.undo", vaccels_undo);
+
 	/* Set up font descriptions for the drawing area */
 	style_context = gtk_widget_get_style_context (hitori->drawing_area);
 	gtk_style_context_get (style_context,
@@ -131,11 +177,41 @@ lookup_color (GtkStyleContext *style_context, const gchar *name, GdkRGBA *result
 	}
 }
 
+/* Generate the text for a given cell, potentially localised to the current locale. */
+static const gchar *
+localise_cell_digit (Hitori             *hitori,
+                     const HitoriVector *pos)
+{
+	guchar value = hitori->board[pos->x][pos->y].num;
+
+	G_STATIC_ASSERT (MAX_BOARD_SIZE < 11);
+
+	switch (value) {
+	/* Translators: This is a digit rendered in a cell on the game board.
+	 * Translate it to your locale’s number system if you wish the game
+	 * board to be rendered in those digits. Otherwise, leave the digits as
+	 * Arabic numerals. */
+	case 1: return C_("Board cell", "1");
+	case 2: return C_("Board cell", "2");
+	case 3: return C_("Board cell", "3");
+	case 4: return C_("Board cell", "4");
+	case 5: return C_("Board cell", "5");
+	case 6: return C_("Board cell", "6");
+	case 7: return C_("Board cell", "7");
+	case 8: return C_("Board cell", "8");
+	case 9: return C_("Board cell", "9");
+	case 10: return C_("Board cell", "10");
+	case 11: return C_("Board cell", "11");
+	default:
+		g_assert_not_reached ();
+	}
+}
+
 static void
 draw_cell (Hitori *hitori, GtkStyleContext *style_context, cairo_t *cr, gdouble cell_size, gdouble x_pos, gdouble y_pos,
            HitoriVector iter)
 {
-	gchar *text;
+	const gchar *text;
 	PangoLayout *layout;
 	gint text_width, text_height;
 	GtkStateFlags state;
@@ -214,7 +290,7 @@ draw_cell (Hitori *hitori, GtkStyleContext *style_context, cairo_t *cr, gdouble
 	cairo_stroke (cr);
 
 	/* Draw the text */
-	text = g_strdup_printf ("%u", hitori->board[iter.x][iter.y].num);
+	text = localise_cell_digit (hitori, &iter);
 	layout = pango_cairo_create_layout (cr);
 
 	pango_layout_set_text (layout, text, -1);
@@ -245,7 +321,6 @@ draw_cell (Hitori *hitori, GtkStyleContext *style_context, cairo_t *cr, gdouble
 
 	pango_cairo_show_layout (cr, layout);
 
-	g_free (text);
 	g_object_unref (layout);
 }
 
@@ -411,6 +486,13 @@ hitori_destroy_cb (GtkWindow *window, Hitori *hitori)
 	hitori_quit (hitori);
 }
 
+static void
+quit_cb (GSimpleAction *action, GVariant *parameters, gpointer user_data)
+{
+	HitoriApplication *self = HITORI_APPLICATION (user_data);
+	hitori_quit (self);
+}
+
 void
 hitori_window_state_event_cb (GtkWindow *window, GdkEventWindowState *event, Hitori *hitori)
 {
@@ -682,7 +764,16 @@ board_size_change_cb (GObject *object, GParamSpec *pspec, gpointer user_data)
 
 	size_str = g_settings_get_string (self->settings, "board-size");
 	size = g_ascii_strtoull (size_str, NULL, 10);
-	hitori_set_board_size (self, size);
-
 	g_free (size_str);
+
+	if (size > MAX_BOARD_SIZE) {
+		GVariant *default_size = g_settings_get_default_value (self->settings, "board-size");
+		g_variant_get (default_size, "s", &size_str);
+		g_variant_unref (default_size);
+		size = g_ascii_strtoull (size_str, NULL, 10);
+		g_free (size_str);
+		g_assert (size <= MAX_BOARD_SIZE);
+	}
+
+	hitori_set_board_size (self, size);
 }
